-- SPDX-FileCopyrightText: 2023 Technical University of Denmark
--
-- SPDX-License-Identifier: BSD-3-Clause

{- |
Module      : Noname
Description : The protocol verifier.
Copyright   : 2023 Technical University of Denmark
License     : BSD-3-Clause
Maintainer  : lpkf@dtu.dk
Stability   : experimental
-}
module Noname (Options (..), defaultOpts, noname, nonameWithOpts) where

-- base
import Control.Monad (when)
import Data.Foldable (msum)
import Data.Functor.Identity (runIdentity)
import Data.Maybe (fromMaybe)
import Data.Version (showVersion)
import System.Exit (exitFailure)
import System.IO (stderr)

-- optparse-applicative
import Options.Applicative

-- text
import Data.Text (Text)
import qualified Data.Text.IO.Utf8 as Text.IO

-- unliftio
import UnliftIO (pooledMapConcurrently)

-- noname
import Noname.Parser
import Noname.Reachability
import Noname.State
import Noname.Verification
import Paths_noname (version)

-- | The running mode for the exploration of reachable states.
data RunningMode
  = -- | Automatic exploration of all reachable states, without any user input.
    Automatic
  | -- | Interactive exploration of one particular trace, with user input.
    Interactive

-- | The options for the CLI.
data Options = Options
  { optInput :: Maybe FilePath
  -- ^ Location of the input for the specification to verify.
  , optOutput :: Maybe FilePath
  -- ^ Location of the output for the verification result.
  , optBound :: Maybe Int
  -- ^ Upper bound on the number of transactions.
  , optMode :: RunningMode
  -- ^ Running mode for the exploration of states.
  , optQuiet :: Bool
  -- ^ Whether to be quiet.
  }

-- | The default options.
defaultOpts :: Options
defaultOpts =
  Options
    { optInput = Just ""
    , optOutput = Nothing
    , optBound = Nothing
    , optMode = Automatic
    , optQuiet = False
    }

{- | The protocol verifier with options given as arguments, returning the output
as 'Text'.
-}
nonameWithOpts :: Options -> IO Text
nonameWithOpts opts = do
  (fp, input) <- case optInput opts of
    Nothing -> do
      input <- Text.IO.getContents
      pure ("stdin", input)
    Just fp -> do
      input <- Text.IO.readFile fp
      pure (fp, input)
  case parseSpecification fp input of
    Left e -> Text.IO.putStrLn e *> exitFailure
    Right (n, state) ->
      let bound = fromMaybe n $ optBound opts
      in  explore opts bound 0 [state]

-- | The parser for the options.
optionsParser :: Parser Options
optionsParser =
  Options
    <$> ( (Just <$> argument str (mconcat [metavar "INPUT", help "Read from INPUT file"]))
            <|> flag' Nothing (mconcat [long "stdin", help "Read from stdin"])
        )
    <*> ( optional . strOption $
            mconcat
              [long "output", short 'o', metavar "OUTPUT", help "Write to OUTPUT file"]
        )
    <*> ( optional . option auto $
            mconcat
              [ long "bound"
              , short 'n'
              , metavar "INT"
              , help "Set bound to INT (override bound specified in the input)"
              ]
        )
    <*> flag
      Automatic
      Interactive
      (mconcat [long "interactive", short 'i', help "Use interactive mode"])
    <*> switch (mconcat [long "quiet", short 'q', help "Whether to be quiet"])

-- | The protocol verifier with options parsed from the CLI.
noname :: IO ()
noname = do
  opts <- execParser parser
  output <- nonameWithOpts opts
  case optOutput opts of
    Nothing -> Text.IO.putStrLn output
    Just f -> Text.IO.writeFile f output
 where
  parser =
    info (optionsParser <**> helper <**> versioner) $
      mconcat
        [ progDesc "Verify an (alpha, beta)-privacy specification"
        , header "noname - formal verification of privacy in security protocols"
        ]
  versioner =
    infoOption
      ( mconcat
          [ "noname "
          , showVersion version
          , "\nCopyright 2023 Technical University of Denmark"
          , "\nLicense BSD-3-Clause <https://spdx.org/licenses/BSD-3-Clause.html>"
          , "\nReport bugs to lpkf@dtu.dk"
          ]
      )
      $ mconcat [long "version", short 'v', help "Show version information", hidden]

{- | Explore the reachable states by executing transactions until the bound is
reached.
-}
explore :: Options -> Int -> Int -> [State] -> IO Text
explore _ _ _ [] = pure "There are no more reachable states."
explore opts bound n states = do
  let ntr = mconcat [" after ", tshow n, " transaction", if n > 1 then "s" else ""]
  when (not $ optQuiet opts) . Text.IO.hPutStrLn stderr $
    mconcat ["Number of states", ntr, ": ", tshow $ length states]
  result <- msum <$> pooledMapConcurrently checkConsistency states
  case result of
    Nothing ->
      if n < bound
        then case optMode opts of
          Automatic ->
            let states' = concat . runIdentity $ traverse executeOne states
            in  explore opts bound (n + 1) states'
          Interactive ->
            explore opts bound (n + 1) . concat =<< traverse executeOne states
        else pure $ mconcat ["Bound reached, no privacy violation found", ntr, "."]
    Just v -> pure $ mconcat ["Privacy violation found", ntr, ".\n", v]
